# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Defines an interface for Abstract writer."""

from __future__ import annotations

import abc
import asyncio
import logging
from typing import Literal

from garf.core.report import GarfReport
from garf.io import formatter
from garf.io.telemetry import tracer

logger = logging.getLogger(__name__)


class AbsWriter(abc.ABC):
  """Base class for defining writers."""

  def __init__(
    self,
    array_handling: Literal['strings', 'arrays'] = 'strings',
    array_separator: str = '|',
    **kwargs,
  ) -> None:
    """Initializes AbsWriter."""
    self.array_handling = array_handling
    self.array_separator = array_separator
    self.kwargs = kwargs

  async def awrite(self, report: GarfReport, destination: str) -> str | None:
    """Writes report to destination."""
    return await asyncio.to_thread(
      self.write, report=report, destination=destination
    )

  @abc.abstractmethod
  def write(self, report: GarfReport, destination: str) -> str | None:
    """Writes report to destination."""

  @tracer.start_as_current_span('format_for_write')
  def format_for_write(self, report: GarfReport) -> GarfReport:
    """Prepares report for writing."""
    array_handling_strategy = formatter.ArrayHandlingStrategy(
      type_=self.array_handling, delimiter=self.array_separator
    )
    return formatter.format_report_for_writing(
      report, [array_handling_strategy]
    )



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for defining writer factory."""

from __future__ import annotations

import enum
import inspect
import sys
from importlib.metadata import entry_points
from typing import Any

from garf.io import exceptions
from garf.io.telemetry import tracer
from garf.io.writers import abs_writer
from opentelemetry import trace


@tracer.start_as_current_span('get_writers')
def _get_writers():
  if sys.version_info.major == 3 and sys.version_info.minor == 9:
    try:
      writers = entry_points()['garf_writer']
    except KeyError:
      writers = []
  else:
    writers = entry_points(group='garf_writer')
  return writers


class StrEnumBase(str, enum.Enum):
  """String enum."""


_writer_options = {writer.name: writer.name for writer in _get_writers()}

WriterOption = enum.Enum(
  'WriterOption',
  _writer_options,
  type=StrEnumBase,
)


class GarfIoWriterError(exceptions.GarfIoError):
  """Writer specific exception."""


@tracer.start_as_current_span('create_writer')
def create_writer(
  writer_option: str | WriterOption, **kwargs: str
) -> type[abs_writer.AbsWriter]:
  """Factory function for creating concrete writer.

  Writer is created based on `writer_option` and possible `kwargs` needed
  to correctly instantiate it.

  Args:
    writer_option: Type of writer.
    kwargs: Any possible arguments needed to instantiate writer.

  Returns:
    Concrete instantiated writer.

  Raises:
    ImportError: When writer specific library is not installed.
    GarfIoError: When incorrect writer option is specified.
  """
  span = trace.get_current_span()
  try:
    WriterOption[writer_option]
  except KeyError as e:
    raise GarfIoWriterError(f'{writer_option} is unknown writer type!') from e
  found_writers = {}
  for writer in _get_writers():
    try:
      if writer.name != writer_option:
        continue
      writer_module = writer.load()
      for name, obj in inspect.getmembers(writer_module):
        if inspect.isclass(obj) and issubclass(obj, abs_writer.AbsWriter):
          found_writers[writer.name] = getattr(writer_module, name)
    except ModuleNotFoundError:
      continue
    except ImportError as e:
      if writer_option == writer.name:
        raise e
      continue
  if concrete_writer := found_writers.get(writer_option):
    span.set_attribute('writer.alias', writer_option)
    for k, v in kwargs.items():
      span.set_attribute(f'{writer_option}.{k}', v)
    return concrete_writer(**kwargs)
  raise GarfIoWriterError(f'Failed to load {writer_option}!')


def setup_writers(
  writers: list[str], writer_parameters: dict[str, Any] | None = None
) -> list[abs_writer.AbsWriter]:
  """Returns list of writer clients."""
  if isinstance(writers, str):
    writers = [writers]
  if writer_parameters is None:
    writer_parameters = {}

  clients = []
  for writer_type in writers:
    writer_client = create_writer(writer_type, **writer_parameters)
    if writer_type in ('bq', 'bigquery'):
      _ = writer_client.create_or_get_dataset()
    if writer_type in ('sheet', 'sheets'):
      writer_client.create_or_get_spreadsheet()
    clients.append(writer_client)
  return clients



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for writing data to JSON."""

from __future__ import annotations

import logging
import os
import pathlib
from typing import Literal, Union

import smart_open
from garf.core import report as garf_report
from garf.io import formatter
from garf.io.telemetry import tracer
from garf.io.writers import file_writer

logger = logging.getLogger(__name__)


class JsonWriter(file_writer.FileWriter):
  """Writes Garf Report to JSON.

  Attributes:
    destination_folder: Destination where JSON files are stored.
  """

  def __init__(
    self,
    destination_folder: Union[
      str, os.PathLike[str], pathlib.Path
    ] = pathlib.Path.cwd(),
    format: Literal['json', 'jsonl'] = 'json',
    **kwargs: str,
  ) -> None:
    """Initializes JsonWriter based on a destination_folder.

    Args:
      destination_folder: A local folder where JSON files are stored.
      format: Format of json file ('json', 'jsonl').
      kwargs: Optional keyword arguments to initialize writer.
    """
    super().__init__(destination_folder=destination_folder, **kwargs)
    self.format = format

  @tracer.start_as_current_span('json.write')
  def write(self, report: garf_report.GarfReport, destination: str) -> str:
    """Writes Garf report to a JSON file.

    Args:
      report: Garf report.
      destination: Base file name report should be written to.

    Returns:
      Base filename where data are written.
    """
    report = self.format_for_write(report)
    file_extension = '.json' if self.format == 'json' else '.jsonl'
    destination = formatter.format_extension(
      destination, new_extension=file_extension
    )
    self.create_dir()
    logger.debug('Writing %d rows of data to %s', len(report), destination)
    output_path = os.path.join(self.destination_folder, destination)
    with smart_open.open(output_path, 'w', encoding='utf-8') as f:
      f.write(report.to_json(output=self.format))
    logger.debug('Writing to %s is completed', output_path)
    return f'[JSON] - at {output_path}'



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Writes GarfReport to BigQuery."""

from __future__ import annotations

import os
from typing import Literal

try:
  import pandas as pd
  import pandas_gbq
  from google.cloud import bigquery
except ImportError as e:
  raise ImportError(
    'Please install garf-io with BigQuery support - `pip install garf-io[bq]`'
  ) from e

import contextlib
import logging

import numpy as np
from garf.core import report as garf_report
from garf.io import exceptions, formatter
from garf.io.telemetry import tracer
from garf.io.writers import abs_writer
from google.cloud import exceptions as google_cloud_exceptions

logger = logging.getLogger(__name__)
logging.getLogger('pandas_gbq').setLevel(logging.WARNING)

_WRITE_DISPOSITION_MAPPING = {
  'WRITE_TRUNCATE': 'replace',
  'WRITE_TRUNCATE_DATA': 'replace',
  'WRITE_APPEND': 'append',
  'WRITE_EMPTY': 'fail',
}


class BigQueryWriterError(exceptions.GarfIoError):
  """BigQueryWriter specific errors."""


class BigQueryWriter(abs_writer.AbsWriter):
  """Writes Garf Report to BigQuery.

  Attributes:
    project: Id of Google Cloud Project.
    dataset: BigQuery dataset to write data to.
    location: Location of a newly created dataset.
    write_disposition: Option for overwriting data.
  """

  def __init__(
    self,
    project: str | None = os.getenv('GOOGLE_CLOUD_PROJECT'),
    dataset: str = 'garf',
    location: str = 'US',
    write_disposition: bigquery.WriteDisposition
    | Literal['append', 'replace', 'fail'] = 'replace',
    **kwargs,
  ):
    """Initializes BigQueryWriter.

    Args:
      project: Id of Google Cloud Project.
      dataset: BigQuery dataset to write data to.
      location: Location of a newly created dataset.
      write_disposition: Option for overwriting data.
      kwargs: Optional keywords arguments.
    """
    super().__init__(**kwargs)
    if not project:
      raise BigQueryWriterError(
        'project is required. Either provide it as project parameter '
        'or GOOGLE_CLOUD_PROJECT env variable.'
      )
    self.project = project
    self.dataset_id = f'{project}.{dataset}'
    self.location = location
    if write_disposition in ('replace', 'append', 'fail'):
      self.write_disposition = write_disposition
    elif isinstance(write_disposition, bigquery.WriteDisposition):
      self.write_disposition = _WRITE_DISPOSITION_MAPPING.get(
        write_disposition.name
      )
    elif _WRITE_DISPOSITION_MAPPING.get(write_disposition.upper()):
      self.write_disposition = _WRITE_DISPOSITION_MAPPING.get(
        write_disposition.upper()
      )
    else:
      raise BigQueryWriterError(
        'Unsupported writer disposition, choose one of: replace, append, fail'
      )
    self._client = None

  def __str__(self) -> str:
    return f'[BigQuery] - {self.dataset_id} at {self.location} location.'

  @property
  def client(self) -> bigquery.Client:
    """Instantiated BigQuery client."""
    if not self._client:
      with tracer.start_as_current_span('bq.create_client'):
        self._client = bigquery.Client(self.project)
    return self._client

  @tracer.start_as_current_span('bq.create_or_get_dataset')
  def create_or_get_dataset(self) -> bigquery.Dataset:
    """Gets existing dataset or create a new one."""
    try:
      bq_dataset = self.client.get_dataset(self.dataset_id)
    except google_cloud_exceptions.NotFound:
      bq_dataset = bigquery.Dataset(self.dataset_id)
      bq_dataset.location = self.location
      with contextlib.suppress(google_cloud_exceptions.Conflict):
        bq_dataset = self.client.create_dataset(bq_dataset, timeout=30)
        logger.info('Created new dataset %s', self.dataset_id)
    return bq_dataset

  @tracer.start_as_current_span('bq.write')
  def write(self, report: garf_report.GarfReport, destination: str) -> str:
    """Writes Garf report to a BigQuery table.

    Args:
      report: Garf report.
      destination: Name of the table report should be written to.

    Returns:
      Name of the table in `dataset.table` format.
    """
    report = self.format_for_write(report)
    destination = formatter.format_extension(destination)
    table = f'{self.dataset_id}.{destination}'
    if not report:
      df = pd.DataFrame(
        data=report.results_placeholder, columns=report.column_names
      ).head(0)
    else:
      df = report.to_pandas()
    df = df.replace({np.nan: None})
    logger.debug('Writing %d rows of data to %s', len(df), destination)
    pandas_gbq.to_gbq(
      dataframe=df,
      destination_table=table,
      if_exists=self.write_disposition,
      progress_bar=False,
    )
    logger.debug('Writing to %s is completed', destination)
    return f'[BigQuery] - at {self.dataset_id}.{destination}'



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for writing data with console."""

from __future__ import annotations

from typing import Literal

import rich
from garf.core import report as garf_report
from garf.io.telemetry import tracer
from garf.io.writers import abs_writer
from rich import console, table


class ConsoleWriter(abs_writer.AbsWriter):
  """Writes reports to standard output.

  Attributes:
    page_size: How many row of report should be written
    type: Type of output ('table', 'json').
  """

  def __init__(
    self,
    page_size: int = 10,
    format: Literal['table', 'json', 'jsonl'] = 'table',
    **kwargs: str,
  ) -> None:
    """Initializes ConsoleWriter.

    Args:
        page_size: How many row of report should be written
        format: Type of output.
        kwargs: Optional parameter to initialize writer.
    """
    super().__init__(**kwargs)
    self.page_size = int(page_size)
    self.format = format

  @tracer.start_as_current_span('console.write')
  def write(self, report: garf_report.GarfReport, destination: str) -> None:
    """Writes Garf report to standard output.

    Args:
      report: Garf report.
      destination: Base file name report should be written to.
    """
    report = self.format_for_write(report)
    if self.format == 'table':
      self._write_rich_table(report, destination)
    elif self.format == 'json':
      print(report.to_json())
    elif self.format == 'jsonl':
      print(report.to_jsonl())

  def _write_rich_table(
    self, report: garf_report.GarfReport, destination: str
  ) -> None:
    query_name = destination.split('/')[-1]
    output_table = table.Table(
      title=f'showing results for query <{query_name}>',
      caption=(
        f'showing rows 1-{min(self.page_size, len(report))} '
        f'out of total {len(report)}'
      ),
      box=rich.box.MARKDOWN,
    )
    for header in report.column_names:
      output_table.add_column(header)
    for i, row in enumerate(report):
      if i < self.page_size:
        output_table.add_row(*[str(field) for field in row])
    console.Console().print(output_table)
